#ifndef LISTA_H
#define LISTA_H

#include <stdlib.h>
#include "IF.h"
#include "smp_inherit.h"


/* ===================================================================
                     NODE DECLARATIONS
   =================================================================== 
 */


// Maybe apply here a configuration file and make the DoubleLinkedNode inherit form this one....
// test this later, verifying if inheritance adds any code overhead or object size (probably not!)


/*
* NodeConfig should have 2 typedefs:
* - ElementType: the type of the element in the list
* - NodeType: the final type of the node being created (this means most specialized one, must
*             take care when using 'inheritance')
*
*/
template <class NodeConfig_>
struct LinkedNode
{

    typedef NodeConfig_ NodeConfig;

private:

    // exports the ElementType (enables static inheritance) 
    typedef NodeConfig::Env ElementType;

    // export node type (enables static inheritance) 
    typedef NodeConfig::ResultType NodeType;
public:
   
    // constructor
    LinkedNode(ElementType const &element_) : element(element_) 
    { 
        next = NULL; 
    }

    // adds after the current node (making the added point to the old 'next' of this)
    void addAfter(ElementType const &element) 
    { 
        NodeType *temporary;
        
        temporary = next;
        next = new NodeType(element);

        next->next = temporary;
    }

    // remove the node after this one
    void removeNext()
    {
        NodeType *temporary = next->next; 
        
        delete next;
        
        next = temporary;
    }

    void linkAfter(NodeType *nextNode)
    {
        next = nextNode;
    }

public:

    ElementType element;
    NodeType *next;
};

/* The uppermost class for this is always LinkedNode.
*
* how to use this subclass in 3 easy steps: 
*  1 - define the tree of inheritance you want to build (you may use one or more 'subclasses' of LinkedNode)
*  2 - declare a struct like the following:
*       template <class ElementType_> 
*       struct YourStructName
*       {
*          typedef ElementType_ ElementType;
*          typedef LastSubclass<SecondLastSubclass<...<LinkedNode<YourStructName> ...> > NodeType;
*       }
*  3 - every time you want to use it, declare a variable: YourStructName<TypeOfElementDesired> variableName;
*         OR create a typedef to easy things up: typedef YourStructName<TypeOfElementDesired> YourNewType;
*
*--------------
*  A simpler way is to use the GenerateSubClass template.
*
*/
template <class BaseNode>
struct DoublyLinkedNode :public BaseNode
{
    typedef BaseNode::NodeConfig NodeConfig;

private:
    typedef NodeConfig::Env ElementType;

    // export node type
    typedef NodeConfig::ResultType NodeType;

public:
    // constructor
    DoublyLinkedNode(ElementType const &element_) : BaseNode(element_) 
    { 
        previous = NULL;
    }

    // redefines addAfter to include the link to the previous
    void addAfter(ElementType const &element) 
    { 
        BaseNode::addAfter(element);
        
        NodeType *nextSucessor = next->next;

        next->previous = this;
        if (nextSucessor)
        {
            nextSucessor->previous = next;
        }
    }

    // redefines removeNext to fix the link to the previous
    void removeNext()
    {
        BaseNode::removeNext();

        if (next)
        {
            next->previous = this;
        }
    }

    // redefines linkAfter to fix the link to the previous
    void linkAfter(NodeType *nextNode)
    {
        BaseNode::linkAfter(nextNode);

        if (nextNode != NULL)
        {
            nextNode->previous = this;
        }
    }

    // bidirectional can have addBefore
    void addBefore(ElementType const &element) 
    { 
        NodeType *temporary = new NodeType(element);
        
        temporary->previous = previous;
        temporary->next = this;

        if (previous != NULL)
        {
            previous->next = temporary;
        }

        previous = temporary;
    }

    // redefines removeNext to fix the link to the previous
    void removePrevious()
    {
        NodeType *temporary = previous;
        NodeType *previousPrevious = previous->previous;
    
        if (previousPrevious)
        {
            previousPrevious->next = this;
        }
        previous = previousPrevious;
    }

    NodeType *previous;
};

// this class adds a method to get the number of followers to this link
// this number is calculated by traversing the rest of the list, don't
// depend on it for performance operations.
template <class BaseNode>
struct LengthNode :public BaseNode
{
    typedef BaseNode::NodeConfig NodeConfig;

private:
    typedef NodeConfig::Env ElementType;

    // export node type
    typedef NodeConfig::ResultType NodeType;

public:
    // constructor
    LengthNode(ElementType const &element_) : BaseNode(element_) 
    { 
        //empty
    }

    int getNumberOfFollowers() const 
    {
        int number = 0;
        NodeType *current = next;

        while (current)
        {
            number++;
            current = current->next;
        }

        return number;
    }
};

/* ===================================================================
                     ITERATOR DECLARATIONS
   ===================================================================
*/

/**
* Iterator that supports only forward iteration through the elements.
*
* It redefines ++ and * to advance the pointer and get the element, respectively.
*
* The paremeter must be a class which exports two typedefs:
*  - Env: the type of the node
*  - ResultType: the final resulting type of the iterator 
*/
template <class IteratorConfig_>
class ForwardIterator
{
public:
    typedef IteratorConfig_ IteratorConfig;

private:
    
    // export the node type
    typedef IteratorConfig::Env NodeType;

    // exports the element type (can be get easily, but makes code cleaner)
    typedef NodeType::NodeConfig::Env ElementType;

    // exports iterator type (enables static inheritance)
    typedef IteratorConfig::ResultType IteratorType;

public:

    // constructor
    ForwardIterator(NodeType *nodeStart) : current(nodeStart) { };
    
    // return the current element on *
    ElementType const & operator*() const 
    {
        return current->element;
    }

    void operator++(int)
    {
        current = current->next;
    }

    bool isValid() const
    {
        return (current) ? true : false;
    }


    NodeType *current;   
};

/**
* Augments a iterator to support bidirectional navigation.
*
* The node type must be bidirectional.
*/
template <class BaseIterator>
class BidirectionalIterator : public BaseIterator 
{
public:
    typedef BaseIterator::IteratorConfig IteratorConfig;

private:
    // export the node type
    typedef IteratorConfig::Env NodeType;

    // exports the element type (can be get easily, but makes code cleaner)
    typedef NodeType::NodeConfig::Env ElementType;
private:
    // exports iterator type (enables static inheritance)
    typedef IteratorConfig::ResultType IteratorType;

public:

    // constructor
    BidirectionalIterator(NodeType *nodeStart) : BaseIterator(nodeStart) { };

    void operator--(int)
    {
        current = current->previous;
    }
};

/**
* Reverses the order of a iterator. 
*
* For this to be possible, the base iterator must be bidirectional.
*/
template <class BidirectionalBaseIterator>
class RevIterator : public BidirectionalBaseIterator
{
public:
    typedef BidirectionalBaseIterator::IteratorConfig IteratorConfig;

private:
    // export the node type
    typedef IteratorConfig::Env NodeType;

    // exports the element type (can be get easily, but makes code cleaner)
    typedef NodeType::NodeConfig::Env ElementType;
private:
    // exports iterator type (enables static inheritance)
    typedef IteratorConfig::ResultType IteratorType;

public:

    // constructor
    RevIterator(NodeType *nodeStart) : BidirectionalBaseIterator(nodeStart) { };

    void operator--(int)
    {
        BidirectionalBaseIterator::operator++(10);
    }

    void operator++(int)
    {
        BidirectionalBaseIterator::operator--(10);
    }
};

/* ===================================================================
                     List DECLARATIONS
   ===================================================================
*/

/**
* This template declares a function for getting the .correct insertion point for a 
* ordered linked list.
*/
template <class ElementType, 
          class NodeType>
struct GetAddPositionOrdered
{
    /*
    * Template function that gets the node after which the element must be inserted.
    */
    template <bool front> // true if the element must be added "first", false if "last"
    static NodeType *getPosition(ElementType const &element, NodeType *first, NodeType *last)
    {
        NodeType *current = first;
        NodeType *previous = NULL;
        if (front)
        {
            // position is before the first equal or larger
            while ((current) && (element < current->element))
            {
                previous = current;
                current = current->next;      
            }
        }
        else
        {
            // position is before the larger (after the ones of same size)
            while ((current) && (element < current->element))
            {
                previous = current;
                current = current->next;      
            }
        }
        return previous;
    }
};

/**
* This template declares a function for getting the .correct insertion point for a 
* linked list.
*/
template <class ElementType, 
          class NodeType>
struct GetAddPositionPlain 
{   
    /*
    * Template function that gets the node after which the element must be inserted.
    */
    template <bool front> // true if the element must be added first, false if last
    static NodeType *getPosition(ElementType const &element, NodeType *first, NodeType *last)
    {
        if (front)
        {
            return NULL; //adds before the first
        }
        else
        {
            return last;
        }
    }
};

/** This template is used to configure the type of the list to be used
* This is basically done by selecting which node and iterator are to be
* used. The node and the iterator must be compatible.
*/
template <class NodeType_, class IteratorType_, bool Ordered_ = false>
struct ListConfig
{
    typedef NodeType_ NodeType; // export the NodeType
    typedef IteratorType_ IteratorType; // export the iterator

    enum { Ordered = Ordered_ }; // export the configuration value
};

// Base class for a linked list
template <class ElementType_, class ListConfig_>
class BaseList
{

public:
    typedef ElementType_ ElementType;

    typedef ListConfig_ ListConfig;

    typedef ListConfig::IteratorType IteratorType;

protected:

    typedef ListConfig::NodeType NodeType;

    enum { Ordered = ListConfig::Ordered };

    typedef IF<Ordered, GetAddPositionOrdered<ElementType, NodeType>, 
                        GetAddPositionPlain<ElementType, NodeType> 
              >::RET AddPosition;

    // export the list type
    typedef BaseList<ElementType, ListConfig> ListType;

public:

    BaseList() : firstNode(NULL), lastNode(NULL) { } // prevents instantiation


    BaseList(NodeType *first) : firstNode(first), lastNode(first) { }

    IteratorType getFirst()
    {   
        IteratorType it(firstNode);

        return it;
    }

    IteratorType getLast()
    {
        IteratorType it(firstNode);

        return it;
    }

    void addFirst(ElementType const &element)
    {
        // inefficient implementation. redundant checks. should implement using a dummy first node
        NodeType *position = AddPosition::getPosition<true>(element, firstNode, lastNode);

        NodeType *newNode = new NodeType(element);

        if (position == NULL)
        {
            NodeType *temporary = firstNode;
            firstNode = newNode;
            
            firstNode->linkAfter(temporary);

            if (lastNode == NULL)
            {
                lastNode = firstNode;
            }
        }
        else
        {
            position->addAfter(element);
            if (lastNode == position)
            {
                lastNode = position->next;
            }
        }
    }

    void removeFirst()
    {
        NodeType *temporary = firstNode->next;

        if (lastNode == firstNode)
        {
            lastNode = temporary; // or null
        }

        delete firstNode;

        firstNode = temporary;
    }

    void addLast(ElementType const &element)
    {
        if (lastNode == NULL)
        {
            addFirst(element);
        }
        else
        {
            // inefficient implementation. redundant checks. should implement using a dummy first node
            NodeType *position = AddPosition::getPosition<false>(element, firstNode, lastNode);

            NodeType *newNode = new NodeType(element);

            if (position == NULL)
            {
                NodeType *temporary = firstNode;
                firstNode = newNode;
            
                firstNode->linkAfter(temporary);

                if (lastNode == NULL)
                {
                    lastNode = firstNode;
                }
            }
            else
            {
                position->addAfter(element);
                if (lastNode == position)
                {
                    lastNode = position->next;
                }
            }
        }
    }

    bool isEmpty()
    {
        return firstNode ? false : true;
    }

    
protected:
    
    NodeType *firstNode;
    NodeType *lastNode;
};

template <class BaseList>
class DoublyAddOn : public BaseList
{
protected:

    typedef RevIterator<IteratorType> RevIteratorType;

public:

    RevIteratorType getFirstRev()
    {
        RevIteratorType it(firstNode);

        return it;
    }
    
    RevIteratorType getLastRev()
    {
        RevIteratorType it(lastNode);

        return it;    
    }    

    void removeLast()
    {
        NodeType *lastPrevious = lastNode->previous;
        if (lastPrevious)
        {
            lastPrevious->removeNext();
            lastNode = lastPrevious;
        }
        else
        {
            delete lastNode;
            firstNode = NULL;
            lastNode = NULL;
        }
    }
};

template <class BaseList>
class StackMutator : protected BaseList
{
protected:
    typedef BaseList::ElementType ElementType;

    typedef BaseList::ListConfig ListConfig;

    typedef ListConfig::IteratorType IteratorType;

public:
    void push(ElementType const &element)
    {
        addFirst(element);
    }

    void pop()
    {
        removeFirst();
    }   
    
    ElementType &peek()
    {
        return firstNode->element;
    } 

    bool isEmpty()
    {
        return BaseList::isEmpty();
    }
};

template <class BaseList>
class DequeMutator : protected BaseList
{
protected:
    typedef BaseList::ElementType ElementType;

    typedef BaseList::ListConfig ListConfig;

    typedef ListConfig::IteratorType IteratorType;

public:
    void pushFront(ElementType const &element)
    {
        addFirst(element);
    }

    void popFront()
    {
        removeFirst();
    }   
    
    void pushBack(ElementType const &element)
    {
        addLast(element);
    }

    void popBack()
    {
        removeLast();
    }   

    ElementType &peekFront()
    {
        return firstNode->element;
    } 

    ElementType &peekBack()
    {
        return lastNode->element;
    } 

    bool isEmpty()
    {
        return BaseList::isEmpty();
    }
};

template <class BaseList>
class QueueMutator : protected BaseList
{
protected:
    typedef BaseList::ElementType ElementType;

    typedef BaseList::ListConfig ListConfig;

    typedef ListConfig::IteratorType IteratorType;

public:
    void push(ElementType const &element)
    {
        addLast(element);
    }

    void pop()
    {
        removeFirst();
    }   
    
    ElementType &peek()
    {
        return firstNode->element;
    } 

    bool isEmpty()
    {
        return BaseList::isEmpty();
    }
};

/* ===================================================================
                     List Flavour PICKER
   ===================================================================
*/

template <class ListDeclareHelper>
class ListFlavourPick
{
private:    
    struct FlavourPublic : public ListDeclareHelper::ListType { };
    struct FlavourProtected : protected ListDeclareHelper::ListType { };

public:

    typedef IF<ListDeclareHelper::publicInherit, FlavourPublic, FlavourProtected>::RET Flavour;
};

/* ===================================================================
                     List Declaration HELPers
   ===================================================================
*/

// linked list
template <class ElementType, bool Ordered>
class LinkedListDeclareHelp
{
private:

    typedef GenerateSubClass<ElementType, LinkedNode>::ResultType NodeType;
    typedef GenerateSubClass<NodeType, ForwardIterator>::ResultType IteratorType;
public:

    typedef BaseList<ElementType, 
                     ListConfig<NodeType, IteratorType, Ordered> 
                     > ListType;

    enum { publicInherit = true };
};

// stack
template <class ElementType, bool Ordered>
class StackDeclareHelp
{
private:
    typedef GenerateSubClass<ElementType, LinkedNode>::ResultType NodeType;
    typedef GenerateSubClass<NodeType, ForwardIterator>::ResultType IteratorType;
public:

    typedef StackMutator<BaseList<ElementType, 
                     ListConfig<NodeType, IteratorType, Ordered> 
                     > > ListType;

    enum { publicInherit = true };
};

// deque
template <class ElementType, bool Ordered>
class DequeDeclareHelp
{
private:
    typedef GenerateSubClass<ElementType, LinkedNode, SubTypeList<DoublyLinkedNode> >::ResultType NodeType;
    typedef GenerateSubClass<NodeType, ForwardIterator, SubTypeList<BidirectionalIterator> >::ResultType IteratorType;

public:

    typedef DequeMutator<DoublyAddOn<BaseList<ElementType, 
                     ListConfig<NodeType, IteratorType, Ordered> 
                     > > > ListType;

    enum { publicInherit = true };
};

// Queue
template <class ElementType, bool Ordered>
class QueueDeclareHelp
{
private:
    typedef GenerateSubClass<ElementType, LinkedNode, SubTypeList<DoublyLinkedNode> >::ResultType NodeType;
    typedef GenerateSubClass<NodeType, ForwardIterator, SubTypeList<BidirectionalIterator> >::ResultType IteratorType;

public:

    typedef QueueMutator<DoublyAddOn<BaseList<ElementType, 
                     ListConfig<NodeType, IteratorType, Ordered> 
                     > > > ListType;

    enum { publicInherit = true };
};

/* ===================================================================
                     Easy to use declarations
   ===================================================================
*/

template <class ElementType, bool Ordered = false>
class LinkedList : public ListFlavourPick<LinkedListDeclareHelp<ElementType, Ordered> >::Flavour 
{ };

template <class ElementType, bool Ordered = false>
class Stack : public ListFlavourPick<StackDeclareHelp<ElementType, Ordered> >::Flavour 
{ };

template <class ElementType, bool Ordered = false>
class Deque : public ListFlavourPick<DequeDeclareHelp<ElementType, Ordered> >::Flavour 
{ };

template <class ElementType, bool Ordered = false>
class Queue : public ListFlavourPick<QueueDeclareHelp<ElementType, Ordered> >::Flavour 
{ };

#endif //LISTA_H
